from google.protobuf.struct_pb2 import Struct
from google.protobuf.empty_pb2 import Empty
from perxis.common import common_pb2
from perxis.items import items_pb2, items_pb2_grpc
from perxis.references import references_pb2
from perxis.references import references_pb2_grpc
from perxis.models import Reference
from perxis.channel import GrpcChannel
from perxis.files import files_pb2, files_pb2_grpc


DEFAULT_PAGE_SIZE: int = 100
DEFAULT_PART_SIZE: int = 1024 * 5


class PerxisDataProvider:
    def __init__(
        self,
        channel: GrpcChannel,
        space_id: str,
        env_id: str,
    ) -> None:
        self.channel = channel
        self.space_id = space_id
        self.env_id = env_id
        self.items_stub = items_pb2_grpc.ItemsStub(self.channel.channel)
        self.references_stub = references_pb2_grpc.ReferencesStub(self.channel.channel)

    async def create(
        self, data: Struct, collection_id: str
    ) -> items_pb2.CreateResponse:
        """Сохранение данных формы в системе Perxis."""
        result = await self.items_stub.Create(
            items_pb2.CreateRequest(
                item=items_pb2.Item(
                    space_id=self.space_id,
                    env_id=self.env_id,
                    collection_id=collection_id,
                    data=data,
                )
            )
        )
        return result

    async def update(self, item_id: str, data: Struct, collection_id: str) -> Empty:
        """Метод обновления записи в коллекции."""
        result = await self.items_stub.Update(
            items_pb2.UpdateRequest(
                item=items_pb2.Item(
                    id=item_id,
                    space_id=self.space_id,
                    env_id=self.env_id,
                    collection_id=collection_id,
                    data=data,
                )
            )
        )
        return result

    async def find(
        self,
        collection_id: str,
        filters: list[str] | None = None,
        sort_by: list[str] | None = None,
        page_num: int | None = None,
        page_size: int | None = DEFAULT_PAGE_SIZE,
    ) -> items_pb2.FindResponse:
        """Метод поиска записей в коллекции."""
        result = await self.items_stub.Find(
            items_pb2.FindRequest(
                space_id=self.space_id,
                env_id=self.env_id,
                collection_id=collection_id,
                filter=items_pb2.Filter(q=filters or []),
                options=items_pb2.FindOptions(
                    options=common_pb2.FindOptions(
                        sort=sort_by, page_num=page_num, page_size=page_size
                    )
                ),
            )
        )
        return result

    async def find_published(
        self,
        collection_id: str,
        filters: list[str] | None = None,
        fields: list[str] | None = None,
        sort_by: list[str] | None = None,
        page_num: int | None = None,
        page_size: int | None = DEFAULT_PAGE_SIZE,
    ) -> items_pb2.FindResponse:
        """Метод поиска опубликованных записей в коллекции."""
        result = await self.items_stub.FindPublished(
            items_pb2.FindPublishedRequest(
                space_id=self.space_id,
                env_id=self.env_id,
                collection_id=collection_id,
                filter=items_pb2.Filter(q=filters or []),
                options=items_pb2.FindPublishedOptions(
                    options=common_pb2.FindOptions(
                        sort=sort_by,
                        page_size=page_size,
                        fields=fields or [],
                        page_num=page_num,
                    )
                ),
            )
        )
        return result

    async def fetch_all_published(
        self,
        collection_id: str,
        filters: list[str] | None = None,
        fields: list[str] | None = None,
        sort_by: list[str] | None = None,
        page_size: int | None = DEFAULT_PAGE_SIZE,
    ) -> items_pb2.FindResponse:
        """Метод поиска всех опубликованных записей в коллекции."""
        kwargs = {
            "collection_id": collection_id,
            "filters": filters,
            "sort_by": sort_by,
            "fields": fields,
            "page_size": page_size,
        }
        message = await self.find_published(**kwargs)
        yield message

        if pages := message.total // page_size:
            if message.total % page_size:
                pages += 1
        else:
            pages = 1

        for page_num in range(1, pages + 1):
            yield await self.find_published(page_num=page_num, **kwargs)

    async def unpublish(self, item_id: str, collection_id: str) -> Empty:
        """Метод снятия с публикации записи в коллекции."""
        result = await self.items_stub.Unpublish(
            items_pb2.UnpublishRequest(
                item=items_pb2.Item(
                    id=item_id,
                    space_id=self.space_id,
                    env_id=self.env_id,
                    collection_id=collection_id,
                )
            )
        )
        return result

    async def publish(self, item_id: str, collection_id: str) -> Empty:
        """Метод публикации записи в коллекции."""
        result = await self.items_stub.Publish(
            items_pb2.PublishRequest(
                item=items_pb2.Item(
                    id=item_id,
                    space_id=self.space_id,
                    env_id=self.env_id,
                    collection_id=collection_id,
                )
            )
        )
        return result

    async def fetch_all(
        self,
        collection_id: str,
        filters: list[str] | str = None,
        sort_by: list[str] | str = None,
        page_size: int | None = DEFAULT_PAGE_SIZE,
    ) -> items_pb2.FindResponse:
        """Метод получения всех записей коллекции."""
        items = []
        storage_data = await self.find(
            collection_id=collection_id,
            page_size=page_size,
            filters=filters,
        )
        items.extend(storage_data.items)

        if pages := storage_data.total // page_size:
            if storage_data.total % page_size:
                pages += 1
        else:
            pages = 1

        for page_num in range(1, pages + 1):
            storage_data = await self.find(
                collection_id=collection_id,
                filters=filters,
                sort_by=sort_by,
                page_num=page_num,
                page_size=page_size,
            )
            items.extend(storage_data.items)
        return items_pb2.FindResponse(items=items, total=len(items))

    async def get_references(
        self, references: list[Reference]
    ) -> items_pb2.GetResponse:
        """Метод получения связей."""
        result = await self.references_stub.Get(
            references_pb2.GetRequest(
                space_id=self.space_id,
                env_id=self.env_id,
                references=[
                    references_pb2.Reference(id=ref.id, collection_id=ref.collection_id)
                    for ref in references
                ],
            )
        )
        return result


class PerxisFileProvider:
    def __init__(self, channel: GrpcChannel) -> None:
        self._stub = files_pb2_grpc.FilesStub(channel.channel)

    async def start_upload(
        self, file_name: str, file_size: int
    ) -> files_pb2.StartUploadResponse:
        message = await self._stub.StartUpload(
            files_pb2.StartUploadRequest(
                upload=files_pb2.MultipartUpload(
                    file=files_pb2.File(
                        name=file_name, size=file_size
                    )
                )
            )
        )
        return message

    async def complete_upload(
        self,
        file_id: str,
        upload_id: str,
        parts: list[str],
        part_size: int = DEFAULT_PART_SIZE
    ) -> files_pb2.CompleteUploadResponse:
        message = await self._stub.CompleteUpload(
            files_pb2.CompleteUploadRequest(
                upload=files_pb2.MultipartUpload(
                    file=files_pb2.File(id=file_id),
                    upload_id=upload_id,
                    part_size=part_size,
                    parts=[
                        files_pb2.CompletedPart(id=value, number=index)
                        for index, value in enumerate(parts, 1)
                    ],

                )
            )
        )
        return message
